from typing import Optional, Tuple, Union

import maccel
import torch
import torch.nn as nn
import numpy as np
import math

from transformers import (
    LlamaPreTrainedModel,
    LlamaConfig,
    LlamaTokenizerFast,
    AutoConfig,
    AutoModel,
    AutoTokenizer,
    AutoModelForCausalLM,
)
from transformers.processing_utils import Unpack
from transformers.modeling_outputs import CausalLMOutputWithPast
from transformers.utils import TransformersKwargs, logging

from mblt_model_zoo.transformers.utils.generation_utils import MobilintGenerationMixin
from ..utils.cache_utils import MobilintCache


logger = logging.get_logger(__name__)


class MobilintLlamaConfig(LlamaConfig):
    model_type = "mobilint-llama"

    def __init__(
        self,
        mxq_path: str = "",
        dev_no: int = 0,
        **kwargs,
    ):
        self.mxq_path = mxq_path
        self.dev_no = dev_no

        super().__init__(**kwargs)

        self.tie_word_embeddings = False


class MobilintLlamaForCausalLM(LlamaPreTrainedModel, MobilintGenerationMixin):
    config: MobilintLlamaConfig
    supports_gradient_checkpointing = False
    _supports_flash_attn = False
    _supports_sdpa = False
    _supports_flex_attn = False

    _can_compile_fullgraph = False
    _supports_attention_backend = False
    _can_record_outputs = {}

    def __init__(self, config: MobilintLlamaConfig, *inputs, **kwargs):
        super().__init__(config, *inputs, **kwargs)
        self.padding_idx = config.pad_token_id
        self.vocab_size = config.vocab_size

        self.embed_tokens = nn.Embedding(
            config.vocab_size, config.hidden_size, self.padding_idx
        )
        self.gradient_checkpointing = False

        self.dev_no = config.dev_no
        self.acc = maccel.Accelerator(self.dev_no)
        mc = maccel.ModelConfig()
        mc.set_single_core_mode(1)
        self.mxq_model = maccel.Model(f"{config.name_or_path}/{config.mxq_path}", mc)
        self.mxq_model.launch(self.acc)
    
    def get_mxq_model(self):
        return self.mxq_model

    def set_decoder(self, decoder):
        raise NotImplementedError("self.model is implemented in mxq")

    def get_decoder(self):
        logger.warning_once("self.model is implemented in mxq")
        return None

    def tie_weights(self):
        pass

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[MobilintCache] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        cache_position: Optional[torch.LongTensor] = None,
        logits_to_keep: Union[int, torch.Tensor] = 0,
        chunk_size: int = 0,
        **kwargs: Unpack[TransformersKwargs],
    ) -> Union[Tuple, CausalLMOutputWithPast]:
        if logits_to_keep > 1:
            logger.warning(
                "logits_to_keep larger than 1 is not supported: %d"
                % logits_to_keep
            )

        if (input_ids is None) ^ (inputs_embeds is not None):
            raise ValueError(
                "You must specify exactly one of input_ids or inputs_embeds"
            )

        if inputs_embeds is None:
            inputs_embeds = self.embed_tokens(input_ids)

        if use_cache and past_key_values is None:
            past_key_values = MobilintCache(self.mxq_model)

        if cache_position is None:
            past_seen_tokens = (
                past_key_values.get_seq_length() if past_key_values is not None else 0
            )
            cache_position = torch.arange(
                past_seen_tokens,
                past_seen_tokens + inputs_embeds.shape[1],
                device=inputs_embeds.device,
            )

        inputs_embeds = inputs_embeds.type(torch.float32).cpu().numpy()

        if inputs_embeds.ndim == 3:
            inputs_embeds = np.expand_dims(
                inputs_embeds, 1
            )  # (batch, 1, seqlen, hidden_size)

        # max width should be appropriate number for chunking (ex. 192 for Llama 3.2 3B)
        # it should be searched experimentally
        if chunk_size == 0:
            chunk_size = self.mxq_model.get_input_buffer_info()[0].max_width
        num_of_chunks = math.ceil(inputs_embeds.shape[2] / chunk_size)

        for i in range(num_of_chunks):
            start_index = i * chunk_size
            end_index = min(start_index + chunk_size, inputs_embeds.shape[2])
            cache_size = (
                0 if past_key_values is None else past_key_values.get_seq_length()
            )

            # last infer
            if i == num_of_chunks - 1:
                logits = self.mxq_model.infer(
                    [inputs_embeds[:, :, start_index:end_index, :]], None, cache_size
                )[0]
            else:
                logits = self.mxq_model.infer(
                    [inputs_embeds[:, :, start_index:end_index, :]], None, cache_size
                )[0]

            if use_cache:
                past_key_values.update_cache_position(
                    cache_position[start_index:end_index]
                )

        logits = torch.tensor(logits, dtype=torch.float32, device=self.device).squeeze(0)

        loss = None
        if labels is not None:
            loss = self.loss_function(
                logits=logits,
                labels=labels,
                vocab_size=self.config.vocab_size,
                **kwargs,
            )

        return CausalLMOutputWithPast(
            loss=loss,
            logits=logits,
            past_key_values=past_key_values,
            hidden_states=None,
            attentions=None,
        )

    def dispose(self):
        self.mxq_model.dispose()


AutoConfig.register("mobilint-llama", MobilintLlamaConfig)
AutoModel.register(MobilintLlamaConfig, MobilintLlamaForCausalLM)
AutoTokenizer.register(MobilintLlamaConfig, fast_tokenizer_class=LlamaTokenizerFast)
AutoModelForCausalLM.register(MobilintLlamaConfig, MobilintLlamaForCausalLM)

from ..utils.types import TransformersModelInfo

Llama_32_1B_Instruct = TransformersModelInfo(
    original_model_id="meta-llama/Llama-3.2-1B-Instruct",
    model_id="mobilint/Llama-3.2-1B-Instruct",
    download_url_base="https://dl.mobilint.com/model/transformers/llm/Llama-3.2-1B-Instruct/",
    file_list=[
        "config.json",
        "generation_config.json",
        "Llama-3.2-1B-Instruct.mxq",
        "model.safetensors",
        "special_tokens_map.json",
        "tokenizer.json",
        "tokenizer_config.json",
    ],
)

Llama_32_3B_Instruct = TransformersModelInfo(
    original_model_id="meta-llama/Llama-3.2-3B-Instruct",
    model_id="mobilint/Llama-3.2-3B-Instruct",
    download_url_base="https://dl.mobilint.com/model/transformers/llm/Llama-3.2-3B-Instruct/",
    file_list=[
        "config.json",
        "generation_config.json",
        "Llama-3.2-3B-Instruct.mxq",
        "model.safetensors",
        "special_tokens_map.json",
        "tokenizer.json",
        "tokenizer_config.json",
    ],
)

Llama_31_8B_Instruct = TransformersModelInfo(
    original_model_id="meta-llama/Llama-3.1-8B-Instruct",
    model_id="mobilint/Llama-3.1-8B-Instruct",
    download_url_base="https://dl.mobilint.com/model/transformers/llm/Llama-3.1-8B-Instruct/",
    file_list=[
        "config.json",
        "generation_config.json",
        "Llama-3.1-8B-Instruct.mxq",
        "model.safetensors",
        "special_tokens_map.json",
        "tokenizer.json",
        "tokenizer_config.json",
    ],
)

HyperCLOVAX_SEED_Text_Instruct_15B = TransformersModelInfo(
    original_model_id="naver-hyperclovax/HyperCLOVAX-SEED-Text-Instruct-1.5B",
    model_id="mobilint/HyperCLOVAX-SEED-Text-Instruct-1.5B",
    download_url_base="https://dl.mobilint.com/model/transformers/llm/HyperCLOVAX-SEED-Text-Instruct-1.5B/",
    file_list=[
        "config.json",
        "HyperCLOVAX-SEED-Text-Instruct-1.5B.mxq",
        "model.safetensors",
        "special_tokens_map.json",
        "tokenizer.json",
        "tokenizer_config.json",
    ],
)



import numpy as np
import torch
import cv2
from typing import List
from .base import PreOps
from ..types import TensorLike


class YoloPre(PreOps):
    """
    Preprocess for YOLO model.
    Original code:
    https://github.com/ultralytics/ultralytics/blob/main/ultralytics/data/augment.py#L1535
    Use default values, where scale fill is False, scale up is True, and auto is False, and center is True.
    """

    def __init__(self, img_size: List[int]):
        super().__init__()
        self.img_size = img_size

    def __call__(self, x: TensorLike):
        if isinstance(x, torch.Tensor):
            x = x.cpu().numpy()
        img = x
        h0, w0 = img.shape[:2]  # original hw
        r = min(self.img_size[0] / h0, self.img_size[1] / w0)  # ratio
        new_unpad = int(round(w0 * r)), int(round(h0 * r))
        dh, dw = (
            self.img_size[0] - new_unpad[1],
            self.img_size[1] - new_unpad[0],
        )  # wh padding

        dw /= 2  # divide padding into 2 sides
        dh /= 2
        if (img.shape[1], img.shape[0]) != new_unpad:
            img = cv2.resize(img, new_unpad, interpolation=cv2.INTER_LINEAR)
        top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
        left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
        img = cv2.copyMakeBorder(
            img, top, bottom, left, right, cv2.BORDER_CONSTANT, value=(114, 114, 114)
        )  # add border
        img = (img / 255).astype(np.float32)

        return torch.from_numpy(img).to(self.device)



import torch
import torch.nn.functional as F
from .base import YOLOPostBase
from .common import *


class YOLOAnchorlessPost(YOLOPostBase):
    def __init__(self, pre_cfg: dict, post_cfg: dict):
        super().__init__(pre_cfg, post_cfg)
        self.make_anchors()
        self.reg_max = 16
        self.no = self.nc + self.reg_max * 4  # number of outputs per anchor (144)
        self.dfl_weight = torch.arange(
            self.reg_max, dtype=torch.float, device=self.device
        ).reshape(1, -1, 1, 1)

    def dfl(self, x):
        assert x.ndim == 3, "Assume that x is a 3d tensor"
        b, _, a = x.shape
        return F.conv2d(
            x.view(b, 4, self.reg_max, a).transpose(2, 1).softmax(1), self.dfl_weight
        ).view(b, 4, a)

    def process_extra(self, x, ic):
        return x

    def make_anchors(self, offset=0.5):
        anchor_points, stride_tensor = [], []
        strides = [2 ** (3 + i) for i in range(self.nl)]
        if self.nl == 2:
            strides = [strd * 2 for strd in strides]

        for strd in strides:
            ny, nx = self.imh // strd, self.imw // strd
            sy = torch.arange(ny, dtype=torch.float32, device=self.device) + offset
            sx = torch.arange(nx, dtype=torch.float32, device=self.device) + offset
            yv, xv = torch.meshgrid(sy, sx, indexing="ij")
            anchor_points.append(torch.stack((xv, yv), -1).reshape(-1, 2))
            stride_tensor.append(
                torch.full((ny * nx, 1), strd, dtype=torch.float32, device=self.device)
            )

        self.anchors = torch.cat(anchor_points, dim=0).permute(1, 0)
        self.stride = torch.cat(stride_tensor, dim=0).permute(1, 0)

    def rearrange(self, x):
        y_det = []
        y_cls = []
        for xi in x:
            if xi.ndim == 3:
                xi = xi.unsqueeze(0)

            assert xi.ndim == 4, f"Got unexpected shape for x={x.shape}."

            if xi.shape[1] == self.reg_max * 4:
                y_det.append(xi)  # (b, 64, 80, 80), (b, 64 ,40, 40), ...
            elif xi.shape[1] == self.nc:
                y_cls.append(xi)  # (b, 80, 80, 80), (b, 80, 40, 40), ...
            else:
                raise ValueError(f"Wrong shape of input: {xi.shape}")

        # sort as box, scores
        y_det = sorted(y_det, key=lambda x: x.numel(), reverse=True)
        y_cls = sorted(y_cls, key=lambda x: x.numel(), reverse=True)

        y = [
            torch.cat((yi_det, yi_cls), dim=1).flatten(2)
            for (yi_det, yi_cls) in zip(y_det, y_cls)
        ]

        return y

    def decode(self, x):
        batch_box_cls = torch.cat(x, axis=-1)  # (b, no=144, 8400)

        y = []
        for xi in batch_box_cls:
            if self.n_extra == 0:
                ic = (torch.amax(xi[-self.nc :, :], dim=0) > self.inv_conf_thres).to(
                    self.device
                )

            else:
                ic = (
                    torch.amax(xi[-self.nc - self.n_extra : -self.n_extra, :], dim=0)
                    > self.inv_conf_thres
                ).to(self.device)

            xi = xi[:, ic]  # (144, *)

            if xi.numel() == 0:
                y.append(
                    torch.zeros(
                        (0, 4 + self.nc + self.n_extra),
                        dtype=torch.float32,
                        device=self.device,
                    )
                )
                continue

            box, score, extra = torch.split(
                xi[None], [self.reg_max * 4, self.nc, self.n_extra], 1
            )  # (1, 64, *), (1, nc, *), (1, n_extra, *)
            dbox = (
                dist2bbox(self.dfl(box), self.anchors[:, ic], xywh=False, dim=1)
                * self.stride[:, ic]
            )
            extra = self.process_extra(extra, ic)
            y.append(
                torch.cat([dbox, score.sigmoid(), extra], dim=1)
                .squeeze(0)
                .transpose(1, 0)
            )

        return y

    def nms(self, x, max_det=300, max_nms=30000, max_wh=7680):
        output = []

        for xi in x:
            if xi.numel() == 0:
                output.append(
                    torch.zeros(
                        (0, 6 + self.n_extra), dtype=torch.float32, device=self.device
                    )
                )
                continue

            box, score, extra = xi[:, :4], xi[:, 4 : 4 + self.nc], xi[:, 4 + self.nc :]
            i, j = (score > self.conf_thres).nonzero(as_tuple=False).T

            xi = torch.cat(
                [box[i], xi[i, j + 4, None], j[:, None].float(), extra[i]], 1
            ).to(self.device)

            if xi.numel() == 0:
                output.append(
                    torch.zeros(
                        (0, 6 + self.n_extra), dtype=torch.float32, device=self.device
                    )
                )
                continue

            xi = xi[torch.argsort(xi[:, 4], descending=True)[:max_nms]]

            # NMS
            c = xi[:, 5:6] * max_wh
            boxes, scores = xi[:, :4] + c, xi[:, 4]
            i = non_max_suppression(boxes, scores, self.iou_thres, max_det)
            output.append(xi[i])

        return output


class YOLOAnchorlessSegPost(YOLOAnchorlessPost):
    def __init__(self, pre_cfg: dict, post_cfg: dict):
        super().__init__(
            pre_cfg,
            post_cfg,
        )

    def __call__(self, x, conf_thres=None, iou_thres=None):
        self.set_threshold(conf_thres, iou_thres)
        x = self.check_input(x)
        x, proto_outs = self.rearrange(x)
        x = self.decode(x)
        x = self.nms(x)
        return self.masking(x, proto_outs)

    def rearrange(self, x):
        y_det = []
        y_cls = []
        y_ext = []
        for xi in x:  # list of bchw outputs
            if xi.ndim == 3:
                xi = xi[None]
            elif xi.ndim == 4:
                pass
            else:
                raise NotImplementedError(f"Got unsupported ndim for input: {xi.ndim}.")

            if xi.shape[1] == self.n_extra:
                y_ext.append(xi)  # (b, 32, 160, 160), (b, 32, 80, 80), ...
            elif xi.shape[1] == self.reg_max * 4:
                y_det.append(xi)  # (b, 64, 80, 80), (b, 64 ,40, 40), ...
            elif xi.shape[1] == self.nc:
                y_cls.append(xi)  # (b, 80, 80, 80), (b, 80, 40, 40), ...
            else:
                raise ValueError(f"Wrong shape of input: {xi.shape}")

        # sort as box, scores
        y_ext = sorted(y_ext, key=lambda x: x.numel(), reverse=True)
        proto = y_ext.pop(0)
        y_det = sorted(y_det, key=lambda x: x.numel(), reverse=True)
        y_cls = sorted(y_cls, key=lambda x: x.numel(), reverse=True)

        assert (
            len(y_cls) == len(y_det) == len(y_ext)
        ), "output arguments are not in a proper form"

        y = [
            torch.cat((yi_det, yi_cls, yi_ext), dim=1).flatten(2)
            for (yi_det, yi_cls, yi_ext) in zip(y_det, y_cls, y_ext)
        ]

        return y, proto


class YOLOAnchorlessPosePost(YOLOAnchorlessPost):
    def __init__(
        self,
        pre_cfg: dict,
        post_cfg: dict,
    ):
        super().__init__(pre_cfg, post_cfg)

    def rearrange(self, x):
        y_det = []
        y_cls = []
        y_kpt = []
        for xi in x:  # list of bchw outputs
            if xi.ndim == 3:
                xi = xi[None]
            elif xi.ndim == 4:
                pass
            else:
                raise NotImplementedError(f"Got unsupported ndim for input: {xi.ndim}.")

            if xi.shape[1] == self.reg_max * 4:
                y_det.append(xi)  # (b, 64, 80, 80), (b, 64 ,40, 40), ...
            elif xi.shape[1] == self.nc:
                y_cls.append(xi)  # (b, 1, 80, 80), (b, 1, 40, 40), ...
            elif xi.shape[1] == self.n_extra:
                y_kpt.append(xi.flatten(2))  # (b, 51, 80, 80), (b, 1, 40, 40), ...
            else:
                raise ValueError(f"Wrong shape of input: {xi.shape}")

        # sort as box, scores
        y_det = sorted(y_det, key=lambda x: x.numel(), reverse=True)
        y_cls = sorted(y_cls, key=lambda x: x.numel(), reverse=True)
        y_kpt = sorted(
            y_kpt, key=lambda x: x.numel(), reverse=True
        )  # (b, 51, 6400), (b, 51, 1600), (b, 51, 400)

        y_tmp = [
            torch.cat((yi_det, yi_cls), dim=1).flatten(2)
            for (yi_det, yi_cls) in zip(
                y_det, y_cls
            )  # (b, 65, 6400), (b, 65, 1600), (b, 65, 400)
        ]
        y = [
            torch.cat((yi_tmp, yi_kpt), dim=1) for (yi_tmp, yi_kpt) in zip(y_tmp, y_kpt)
        ]  # (b, 116, 6400), (b, 116, 1600), (b, 116, 400)

        return y

    def process_extra(self, kpt, ic):
        kpt = kpt.squeeze(0)
        assert kpt.shape[0] == self.n_extra, "keypoint shape mismatch"

        kpt = kpt.reshape(17, 3, -1)
        coord, conf = torch.split(kpt, [2, 1], dim=1)  # (17, 2, *), (17, 1, *)

        coord = (coord * 2 + (self.anchors[:, ic] - 0.5)) * self.stride[:, ic]
        conf = conf.sigmoid()
        kpt = torch.cat([coord, conf], dim=1).reshape(self.n_extra, -1)
        return kpt.unsqueeze(0)



from typing import Optional, Tuple, TypeVar, Union, List, Any

import maccel
import torch
import numpy as np

from transformers import (
    PretrainedConfig,
    PreTrainedModel,
    BlipTextConfig,
    BlipVisionConfig,
    BlipPreTrainedModel,
    AutoConfig,
    AutoTokenizer,
    BertTokenizer,
    BertTokenizerFast,
    AutoProcessor,
    BlipProcessor,
    AutoModelForVision2Seq,
    AutoModelForImageTextToText,
)
from transformers.models.blip.modeling_blip import (
    BlipForConditionalGenerationModelOutput,
)
from transformers.models.blip.modeling_blip_text import (
    BlipTextPreTrainedModel,
    BlipTextEmbeddings,
)
from transformers.modeling_outputs import (
    BaseModelOutputWithPoolingAndCrossAttentions,
    CausalLMOutputWithCrossAttentions,
    BaseModelOutputWithPooling,
)
from transformers.utils import logging

from mblt_model_zoo.transformers.utils.generation_utils import MobilintGenerationMixin
from ..utils.cache_utils import MobilintCache


logger = logging.get_logger(__name__)


SpecificPreTrainedModelType = TypeVar(
    "SpecificPreTrainedModelType", bound="PreTrainedModel"
)

# type hinting: specifying the type of config class that inherits from PretrainedConfig
SpecificPretrainedConfigType = TypeVar("SpecificPretrainedConfigType", bound="PretrainedConfig")

class MobilintBlipTextConfig(BlipTextConfig):
    model_type = "mobilint-blip_text_model"

    def __init__(
        self,
        mxq_path: str = "",
        dev_no: int = 0,
        **kwargs,
    ):
        self.mxq_path = mxq_path
        self.dev_no = dev_no

        super().__init__(**kwargs)

        self.tie_word_embeddings = False


class MobilintBlipVisionConfig(BlipVisionConfig):
    model_type = "mobilint-blip_vision_model"

    def __init__(
        self,
        mxq_path: str = "",
        dev_no: int = 0,
        **kwargs,
    ):
        self.mxq_path = mxq_path
        self.dev_no = dev_no

        super().__init__(**kwargs)

        self.tie_word_embeddings = False


class MobilintBlipConfig(PretrainedConfig):
    model_type = "mobilint-blip"
    sub_configs = {
        "text_config": MobilintBlipTextConfig,
        "vision_config": MobilintBlipVisionConfig,
    }

    def __init__(
        self,
        text_config=None,
        vision_config=None,
        projection_dim=512,
        logit_scale_init_value=2.6592,
        image_text_hidden_size=256,
        label_smoothing=0.0,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if text_config is None:
            text_config = {}
            logger.info(
                "`text_config` is `None`. Initializing the `MobilintBlipTextConfig` with default values."
            )

        if vision_config is None:
            vision_config = {}
            logger.info(
                "`vision_config` is `None`. Initializing the `MobilintBlipVisionConfig` with default values."
            )

        self.text_config = MobilintBlipTextConfig(**text_config)
        self.vision_config = MobilintBlipVisionConfig(**vision_config)

        self.text_config.encoder_hidden_size = self.vision_config.hidden_size

        self.projection_dim = projection_dim
        self.logit_scale_init_value = logit_scale_init_value
        self.initializer_factor = 1.0
        self.initializer_range = 0.02
        self.image_text_hidden_size = image_text_hidden_size
        self.label_smoothing = label_smoothing

    @classmethod
    def from_text_vision_configs(
        cls,
        text_config: MobilintBlipTextConfig,
        vision_config: MobilintBlipVisionConfig,
        **kwargs,
    ):
        return cls(
            text_config=text_config.to_dict(),
            vision_config=vision_config.to_dict(),
            **kwargs,
        )
    
    @classmethod
    def from_dict(
        cls: type[SpecificPretrainedConfigType], config_dict: dict[str, Any], **kwargs
    ) -> SpecificPretrainedConfigType:
        return_unused_kwargs = kwargs.pop("return_unused_kwargs", False)
        
        config, kwargs = super().from_dict(config_dict, return_unused_kwargs=True, **kwargs)
        
        config.text_config.name_or_path = config.name_or_path
        config.vision_config.name_or_path = config.name_or_path
        
        if return_unused_kwargs:
            return config, kwargs
        else:
            return config
        

class MobilintBlipPreTrainedModel(BlipPreTrainedModel):
    config_class = MobilintBlipConfig
    supports_gradient_checkpointing = False


class MobilintBlipTextPreTrainedModel(BlipTextPreTrainedModel):
    config_class = MobilintBlipTextConfig


class MobilintBlipTextModel(MobilintBlipTextPreTrainedModel):
    def __init__(self, config: MobilintBlipTextConfig):
        super().__init__(config)
        self.config = config

        self.embeddings = BlipTextEmbeddings(config)

        self.dev_no = config.dev_no
        self.acc = maccel.Accelerator(self.dev_no)
        mc = maccel.ModelConfig()
        mc.set_single_core_mode(
            None, [maccel.CoreId(maccel.Cluster.Cluster1, maccel.Core.Core1)]
        )
        self.mxq_model = maccel.Model(f"{config.name_or_path}/{config.mxq_path}", mc)
        self.mxq_model.launch(self.acc)

    def get_input_embeddings(self):
        return self.embeddings.word_embeddings

    def set_input_embeddings(self, value):
        self.embeddings.word_embeddings = value

    def _prune_heads(self, heads_to_prune):
        raise NotImplementedError("_prune_heads is not implemented")

    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        encoder_embeds: Optional[torch.Tensor] = None,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.Tensor] = None,
        past_key_values: Optional[MobilintCache] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        is_decoder: Optional[bool] = False,
        cache_position: Optional[torch.Tensor] = None,
    ) -> Union[Tuple[torch.Tensor], BaseModelOutputWithPoolingAndCrossAttentions]:
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        if is_decoder:
            use_cache = use_cache if use_cache is not None else self.config.use_cache
        else:
            use_cache = False

        if head_mask is not None:
            logger.warning_once("head_mask is not supported.")

        if encoder_attention_mask is not None:
            logger.warning_once("encoder_attention_mask is not supported.")

        if output_attentions:
            logger.warning_once("output_attentions is not supported.")

        if output_hidden_states:
            logger.warning_once("output_hidden_states is not supported.")

        if input_ids is not None and inputs_embeds is not None:
            raise ValueError(
                "You cannot specify both input_ids and inputs_embeds at the same time"
            )
        elif input_ids is not None:
            self.warn_if_padding_and_no_attention_mask(input_ids, attention_mask)
        else:
            raise ValueError(
                "You have to specify either input_ids or inputs_embeds or encoder_embeds"
            )

        if use_cache or past_key_values is not None:
            if past_key_values is None:
                past_key_values = MobilintCache(self.mxq_model)
            elif not isinstance(past_key_values, MobilintCache):
                logger.warning_once(
                    "Class of past_key_values should be MobilintCache, current: "
                    + past_key_values.__class__.__name__
                )
                past_key_values = MobilintCache(self.mxq_model)

        # past_key_values_length
        past_key_values_length = (
            past_key_values.get_seq_length() if past_key_values is not None else 0
        )

        if encoder_embeds is None:
            embedding_output = self.embeddings(
                input_ids=input_ids,
                position_ids=position_ids,
                inputs_embeds=inputs_embeds,
                past_key_values_length=past_key_values_length,
            )
        else:
            embedding_output = encoder_embeds

        encoder_hidden_states = (
            encoder_hidden_states.unsqueeze(1).type(torch.float32).cpu().numpy()
        )
        embedding_output = (
            embedding_output.unsqueeze(1).type(torch.float32).cpu().numpy()
        )

        logits = self.mxq_model.infer(
            [encoder_hidden_states, embedding_output],
            cache_size=int(past_key_values_length),
        )[0]
        logits = torch.tensor(logits, dtype=torch.float32, device=self.device).squeeze(0)

        past_key_values.update_cache_position(cache_position)

        if not return_dict:
            return (logits, past_key_values)

        return BaseModelOutputWithPoolingAndCrossAttentions(
            last_hidden_state=logits,
            pooler_output=None,
            past_key_values=past_key_values,
            hidden_states=None,
            attentions=None,
            cross_attentions=None,
        )

    def dispose(self):
        self.mxq_model.dispose()


class MobilintBlipTextLMHeadModel(MobilintBlipTextPreTrainedModel, MobilintGenerationMixin):
    def __init__(self, config: MobilintBlipTextConfig):
        super().__init__(config)

        self.bert = MobilintBlipTextModel(config)
        self.label_smoothing = config.label_smoothing
    
    def get_mxq_model(self):
        return self.bert.mxq_model

    def get_input_embeddings(self):
        return self.bert.get_input_embeddings()

    def set_input_embeddings(self, new_embeddings):
        self.bert.set_input_embeddings(new_embeddings)

    def get_output_embeddings(self):
        logger.warning_once("get_output_embeddings is not implemented")
        return None

    def set_output_embeddings(self, new_embeddings):
        raise NotImplementedError("set_output_embeddings is not implemented")

    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.Tensor] = None,
        labels: Optional[torch.Tensor] = None,
        past_key_values: Optional[List[torch.Tensor]] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        return_logits: Optional[bool] = False,
        is_decoder: Optional[bool] = True,
        reduction: Optional[str] = "mean",
        cache_position: Optional[torch.Tensor] = None,
    ) -> Union[Tuple[torch.Tensor], CausalLMOutputWithCrossAttentions]:
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )
        if labels is not None:
            use_cache = False

        if reduction != "mean":
            logger.warning_once(
                "reduction except 'mean' is not supported: " + reduction
            )

        outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            is_decoder=is_decoder,
            cache_position=cache_position,
        )

        prediction_scores = outputs[0]

        if return_logits:
            return prediction_scores[:, :-1, :].contiguous()

        if not return_dict:
            return outputs

        return CausalLMOutputWithCrossAttentions(
            loss=None,
            logits=prediction_scores,
            past_key_values=outputs.past_key_values,
            hidden_states=None,
            attentions=None,
            cross_attentions=None,
        )

    def prepare_inputs_for_generation(self, input_ids, past_key_values=None, attention_mask=None, **model_kwargs):
        # Overwrite -- hardcoded key return (`is_decoder=True`)

        model_inputs = super().prepare_inputs_for_generation(
            input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            **model_kwargs,
        )
        model_inputs["is_decoder"] = True

        return model_inputs

    def dispose(self):
        self.bert.dispose()


class MobilintBlipVisionModel(MobilintBlipPreTrainedModel):
    main_input_name = "pixel_values"
    config_class = MobilintBlipVisionConfig

    def __init__(self, config: MobilintBlipVisionConfig):
        super().__init__(config)
        self.config = config

        self.dev_no = config.dev_no
        self.acc = maccel.Accelerator(self.dev_no)
        mc = maccel.ModelConfig()
        mc.set_global_core_mode([maccel.Cluster.Cluster0])
        self.mxq_model = maccel.Model(f"{config.name_or_path}/{config.mxq_path}", mc)
        self.mxq_model.launch(self.acc)

    def forward(
        self,
        pixel_values: Optional[torch.FloatTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        interpolate_pos_encoding: bool = False,
    ) -> Union[Tuple, BaseModelOutputWithPooling]:
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        if pixel_values is None:
            raise ValueError("You have to specify pixel_values")

        if output_attentions:
            logger.warning_once("output_attentions is not supported.")

        if output_hidden_states:
            logger.warning_once("output_hidden_states is not supported.")

        if interpolate_pos_encoding is True:
            logger.warning_once("interpolate_pos_encoding is not supported.")

        last_hidden_state = self.mxq_model.infer(
            pixel_values.type(torch.float32).cpu().numpy()
        )[0]
        last_hidden_state = np.transpose(last_hidden_state[:, :, 0], (0, 2, 1))
        last_hidden_state = torch.tensor(last_hidden_state, dtype=torch.float32, device=pixel_values.device)

        if not return_dict:
            return (last_hidden_state,)

        return BaseModelOutputWithPooling(
            last_hidden_state=last_hidden_state,
            pooler_output=None,
            hidden_states=None,
            attentions=None,
        )

    def get_input_embeddings(self):
        logger.warning_once("get_input_embeddings is not implemented")
        return None

    def dispose(self):
        self.mxq_model.dispose()


class MobilintBlipForConditionalGeneration(
    MobilintBlipPreTrainedModel, MobilintGenerationMixin
):
    config_class = MobilintBlipConfig
    _tied_weights_keys = []
    main_input_name = "pixel_values"

    def __init__(self, config: MobilintBlipConfig, *inputs, **kwargs):
        super().__init__(config, *inputs, **kwargs)
        
        self.vision_model = MobilintBlipVisionModel(config.vision_config)

        self.text_decoder = MobilintBlipTextLMHeadModel(config.text_config)

        self.decoder_input_ids = config.text_config.bos_token_id
        self.decoder_pad_token_id = config.text_config.pad_token_id
    
    def get_mxq_model(self):
        return self.text_decoder.get_mxq_model()

    def get_input_embeddings(self):
        return self.text_decoder.get_input_embeddings()

    def set_input_embeddings(self, value):
        self.text_decoder.set_input_embeddings(value)

    def forward(
        self,
        pixel_values: torch.FloatTensor,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.LongTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        labels: Optional[torch.LongTensor] = None,
        return_dict: Optional[bool] = None,
        interpolate_pos_encoding: bool = False,
    ) -> Union[Tuple, BlipForConditionalGenerationModelOutput]:
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )

        vision_outputs = self.vision_model(
            pixel_values=pixel_values,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            interpolate_pos_encoding=interpolate_pos_encoding,
        )

        image_embeds = vision_outputs[0]

        outputs = self.text_decoder(
            input_ids=input_ids,
            attention_mask=attention_mask,
            encoder_hidden_states=image_embeds,
            labels=labels,
            return_dict=return_dict,
            reduction="mean",
        )

        if not return_dict:
            outputs += vision_outputs
            return tuple(output for output in outputs if output is not None)

        return BlipForConditionalGenerationModelOutput(
            loss=outputs.loss,
            logits=outputs.logits,
            image_embeds=image_embeds,
            last_hidden_state=vision_outputs.last_hidden_state,
            hidden_states=vision_outputs.hidden_states,
            attentions=vision_outputs.attentions,
        )

    def generate(
        self,
        pixel_values: torch.FloatTensor,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.LongTensor] = None,
        interpolate_pos_encoding: bool = False,
        **generate_kwargs,
    ) -> torch.LongTensor:
        batch_size = pixel_values.shape[0]
        vision_outputs = self.vision_model(
            pixel_values=pixel_values,
            interpolate_pos_encoding=interpolate_pos_encoding,
        )

        image_embeds = vision_outputs[0]

        if isinstance(input_ids, list):
            input_ids = torch.LongTensor(input_ids)
        elif input_ids is None:
            input_ids = (
                torch.LongTensor(
                    [[self.decoder_input_ids, self.config.text_config.eos_token_id]]
                )
                .repeat(batch_size, 1)
                .to(image_embeds.device)
            )

        input_ids[:, 0] = self.config.text_config.bos_token_id

        outputs = self.text_decoder.generate(
            input_ids=input_ids[:, :-1],
            eos_token_id=self.config.text_config.sep_token_id,
            pad_token_id=self.config.text_config.pad_token_id,
            attention_mask=None,
            encoder_hidden_states=image_embeds,
            encoder_attention_mask=None,
            **generate_kwargs,
        )

        return outputs

    def dispose(self):
        self.vision_model.dispose()
        self.text_decoder.dispose()


AutoConfig.register("mobilint-blip", MobilintBlipConfig)
AutoTokenizer.register(MobilintBlipConfig, BertTokenizer, BertTokenizerFast)
AutoProcessor.register(MobilintBlipConfig, BlipProcessor)
AutoModelForVision2Seq.register(
    MobilintBlipConfig, MobilintBlipForConditionalGeneration
)
AutoModelForImageTextToText.register(
    MobilintBlipConfig, MobilintBlipForConditionalGeneration
)

from ..utils.types import TransformersModelInfo

blip_image_captioning_large = TransformersModelInfo(
    original_model_id="Salesforce/blip-image-captioning-large",
    model_id="mobilint/blip-image-captioning-large",
    download_url_base="https://dl.mobilint.com/model/transformers/vlm/blip-image-captioning-large/",
    file_list=[
        "blip-image-captioning-large_text_decoder.mxq",
        "blip-image-captioning-large_vision_model.mxq",
        "config.json",
        "model.safetensors",
        "preprocessor_config.json",
        "special_tokens_map.json",
        "tokenizer.json",
        "tokenizer_config.json",
        "vocab.txt",
    ],
)



from typing import Union, List, Dict
import importlib
import inspect
from .utils import TransformersModelInfo

TASKS = [
    "large_language_model",
    "speech_to_text",
    "text_to_speech",
    "vision_language_model",
]


def list_tasks():
    return TASKS


def list_models(
    tasks: Union[str, List[str]] = TASKS,
) -> Dict[str, TransformersModelInfo]:
    if isinstance(tasks, str):
        tasks = [tasks]
    assert set(tasks).issubset(TASKS), f"mblt model zoo supports tasks in {TASKS}"

    available_models = {}
    for task in tasks:
        available_models[task] = []
        try:
            module = importlib.import_module(
                f".{task}", package=__name__.replace("._api", "")
            )
        except ImportError as e:
            print(f"Failed to import module for task '{task}': {e}")
            continue

        for name, obj in inspect.getmembers(module):
            if (
                not inspect.isclass(obj)
                and isinstance(obj, TransformersModelInfo)
                and obj is not TransformersModelInfo
            ):
                available_models[task].append(obj)

    return available_models



import torch
from .base import YOLOPostBase
from .common import *


class YOLOAnchorPost(YOLOPostBase):
    def __init__(self, pre_cfg: dict, post_cfg: dict):
        super().__init__(pre_cfg, post_cfg)
        self.no = self.nc + 5 + self.n_extra
        self.make_anchor_grid()

    def make_anchor_grid(self):
        self.grid, self.anchor_grid, self.stride = [], [], []

        strides = [2 ** (3 + i) for i in range(self.nl)]
        if self.nl == 2:
            strides = [strd * 2 for strd in strides]

        out_sizes = [
            [self.imh // strd, self.imw // strd] for strd in strides
        ]  # (80, 80), (40, 40), (20, 20)
        for anchr, (ny, nx), strd in zip(self.anchors, out_sizes, strides):
            yv, xv = torch.meshgrid(
                torch.arange(ny, dtype=torch.float32, device=self.device),
                torch.arange(nx, dtype=torch.float32, device=self.device),
                indexing="ij",
            )
            grid = torch.stack((xv, yv), 2).expand(self.na, ny, nx, 2)
            self.grid.append(grid)

            anchr = torch.broadcast_to(
                torch.tensor(anchr).reshape(self.na, 1, 1, 2),
                (self.na, ny, nx, 2),
            ).to(self.device)
            self.anchor_grid.append(anchr)

            self.stride.append(strd * torch.ones(self.na, ny, nx, 2).to(self.device))

        self.grid = torch.cat([grd.reshape(-1, 2) for grd in self.grid], dim=0).to(
            self.device
        )
        self.anchor_grid = torch.cat(
            [anc.reshape(-1, 2) for anc in self.anchor_grid], dim=0
        ).to(self.device)
        self.stride = torch.cat(
            [strd.reshape(-1, 2) for strd in self.stride], dim=0
        ).to(self.device)

    def rearrange(self, x):
        y = []
        for xi in x:
            if xi.ndim == 3:
                xi = xi.unsqueeze(0)

            assert (
                xi.ndim == 4 and xi.shape[1] == self.no * self.na
            ), f"Got unexpected shape for x={xi.shape}."

            y.append(xi)

        y = sorted(
            y, key=lambda x: x.numel(), reverse=True
        )  # sort by number of elements in descending order
        return y

    def decode(self, x):
        x = torch.cat(
            [
                xi.reshape(xi.shape[0], self.na, self.no, xi.shape[2], xi.shape[3])
                .permute(0, 1, 3, 4, 2)
                .reshape(xi.shape[0], -1, self.no)
                for xi in x
            ],
            dim=1,
        ).to(self.device)
        x = list(x.unbind())  # convert to list of tensors if shape is (25200, 85 + 32)

        y = []
        for xi in x:
            ic = (xi[:, 4] > self.inv_conf_thres).to(self.device)  # candidate indices
            box_cls = xi[ic]  # candidate boxes
            if box_cls.numel() == 0:
                y.append(
                    torch.zeros((0, self.no), dtype=torch.float32, device=self.device)
                )
                continue
            grid = self.grid[ic]
            anchor_grid = self.anchor_grid[ic]
            stride = self.stride[ic]

            xy, wh, conf, scores, extra = torch.split(
                box_cls, [2, 2, 1, self.nc, self.n_extra], dim=-1
            )

            xy = (xy.sigmoid() * 2.0 - 0.5 + grid) * stride
            wh = (wh.sigmoid() * 2) ** 2 * anchor_grid
            conf = conf.sigmoid()
            scores = scores.sigmoid()
            extra *= conf
            y.append(torch.cat([xy, wh, conf, scores, extra], dim=-1).to(self.device))
        return y

    def nms(self, x, max_det=300, max_nms=30000, max_wh=7680):
        mi = 5 + self.nc  # mask index
        output = []
        for xi in x:
            if xi.numel() == 0:
                output.append(
                    torch.zeros(
                        (0, 6 + self.n_extra), dtype=torch.float32, device=self.device
                    )
                )
                continue
            xi[..., 5:] *= xi[..., 4:5]  # conf = obj_conf * cls_conf

            box = xywh2xyxy(xi[..., :4])
            mask = xi[..., mi:]

            i, j = (xi[..., 5:mi] > self.conf_thres).nonzero(as_tuple=False).T
            xi = torch.cat(
                [box[i], xi[i, j + 5, None], j[:, None].float(), mask[i]], 1
            ).to(self.device)

            if xi.numel() == 0:
                output.append(torch.zeros((0, 6 + self.n_extra), dtype=torch.float32))
                continue

            xi = xi[xi[:, 4].argsort(descending=True)[:max_nms]]  # sort by confidence

            # NMS
            c = xi[:, 5:6] * max_wh
            boxes, scores = xi[:, :4] + c, xi[:, 4]
            i = non_max_suppression(boxes, scores, self.iou_thres, max_det)
            output.append(xi[i])

        return output


class YOLOAnchorSegPost(YOLOAnchorPost):
    def __init__(self, pre_cfg: dict, post_cfg: dict):
        super().__init__(pre_cfg, post_cfg)

    def __call__(self, x, conf_thres=None, iou_thres=None):
        self.set_threshold(conf_thres, iou_thres)
        x = self.check_input(x)
        x, proto_outs = self.rearrange(x)
        x = self.decode(x)
        x = self.nms(x)
        return self.masking(x, proto_outs)

    def rearrange(self, x):
        proto_exist = False
        for i, xi in enumerate(x):
            if xi.ndim == 3:
                xi = xi.unsqueeze(0)
            assert xi.ndim == 4, f"Got unexpected shape for x={xi.shape}."
            if self.n_extra == xi.shape[1]:
                proto = xi  # (bs, 32, 160, 160)
                x.pop(i)
                proto_exist = True
                break

        if not proto_exist:
            raise ValueError("Proto output is missing.")

        y = super().rearrange(x)
        return y, proto



from ..utils.types import ModelInfo, ModelInfoSet
from ..wrapper import MBLT_Engine


class FlexiViT_Small_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/flexivit_small/aries/single/flexivit_small.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/flexivit_small/aries/multi/flexivit_small.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/flexivit_small/aries/global/flexivit_small.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 252,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [240, 240],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


class FlexiViT_Base_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/flexivit_base/aries/single/flexivit_base.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/flexivit_base/aries/multi/flexivit_base.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/flexivit_base/aries/global/flexivit_base.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 252,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [240, 240],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


class FlexiViT_Large_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/flexivit_large/aries/single/flexivit_large.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/flexivit_large/aries/multi/flexivit_large.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/flexivit_large/aries/global/flexivit_large.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 252,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [240, 240],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


def FlexiViT_Small(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        FlexiViT_Small_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def FlexiViT_Base(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        FlexiViT_Base_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def FlexiViT_Large(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        FlexiViT_Large_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )



from ..utils.types import ModelInfo, ModelInfoSet
from ..wrapper import MBLT_Engine


class YOLOv8xPose_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/pose_estimation/yolov8x-pose/aries/single/yolov8x-pose.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/pose_estimation/yolov8x-pose/aries/multi/yolov8x-pose.mxq",
                    "global": "https://dl.mobilint.com/model/vision/pose_estimation/yolov8x-pose/aries/global/yolov8x-pose.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [640, 640],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "pose_estimation",
            "nc": 1,  # Number of classes
            "nl": 3,  # Number of detection layers
            "n_extra": 51,
        },
    )
    DEFAULT = COCO_V1


def YOLOv8xPose(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv8xPose_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )



import numpy as np
import torch
from .base import PreOps
from ..types import TensorLike


class SetOrder(PreOps):
    """Set the channel order of the image."""

    def __init__(self, shape: str = "HWC"):
        """Set the channel order of the image.

        Args:
            shape (str, optional): Channel order. Defaults to "HWC".
        """
        super().__init__()
        assert shape.lower() in ["hwc", "chw"], f"Got unsupported shape={shape}."
        self.shape = shape

    def __call__(self, x: TensorLike):
        assert x.ndim == 3, "Assume that x is a color image"
        if x.shape[0] == 3:
            cdim = 0
        elif x.shape[-1] == 3:
            cdim = 2
        else:
            raise ValueError(
                f"Only assume HWC or CHW with 3 channels, but got shape {x.shape}"
            )

        is_tensor = isinstance(x, torch.Tensor)
        permute_fn = torch.permute if is_tensor else np.transpose

        if cdim == 0 and self.shape.lower() == "hwc":
            return permute_fn(x, (1, 2, 0))
        elif cdim == 2 and self.shape.lower() == "chw":
            return permute_fn(x, (2, 0, 1))

        return x



from typing import Union, List, Any
import torch
import numpy as np
from dataclasses import dataclass
from collections import OrderedDict
from enum import Enum

TensorLike = Union[torch.Tensor, np.ndarray]
ListTensorLike = List[TensorLike]


@dataclass
class ModelInfo:
    """
    This class is used to store model information.
    """

    pre_cfg: OrderedDict
    post_cfg: OrderedDict
    model_cfg: OrderedDict


class ModelInfoSet(Enum):
    """
    This class is used to store model informations.
    """



from typing import Dict
import torch
import maccel
from abc import ABC, abstractmethod
from transformers import Cache, GenerationConfig, GenerationMixin, PreTrainedModel
from mblt_model_zoo.transformers.utils.cache_utils import MobilintCache

class MobilintGenerationMixin(ABC, GenerationMixin):        
    @abstractmethod
    def get_mxq_model(self) -> maccel.Model:
        pass
    
    # Function arguments changed for transformers>=4.56.0
    # args contain device and model_kwargs in transformers<4.56.0
    # args contain only model_kwargs in transformers>=4.56.0
    def _get_cache(
        self, cache_implementation: str, batch_size: int, max_cache_len: int, *args
    ) -> Cache:
        if not hasattr(self, "_cache"):
            self._cache = MobilintCache(self.get_mxq_model())
        else:
            self._cache.reset()
            
        return self._cache

    # Function arguments changed for transformers>=4.56.0
    # args contain device in transformers<4.56.0
    # args empty in transformers>=4.56.0
    def _prepare_cache_for_generation(
        self,
        generation_config: GenerationConfig,
        model_kwargs: Dict,
        assistant_model: "PreTrainedModel",
        batch_size: int,
        max_cache_length: int,
        *args,
    ) -> bool:
        super()._prepare_cache_for_generation(
            generation_config,
            model_kwargs,
            assistant_model,
            batch_size,
            max_cache_length,
            *args,
        )

        cache_name = "past_key_values"

        if model_kwargs.get(cache_name, None) is None:
            return
        elif model_kwargs[cache_name].__class__.__name__ == "MobilintCache":
            return
        else:
            model_kwargs[cache_name] = self._get_cache("mobilint", batch_size, max_cache_length, *args, model_kwargs)


from ..utils.types import ModelInfo, ModelInfoSet
from ..wrapper import MBLT_Engine


class YOLOv7_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov7/aries/single/yolov7.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov7/aries/multi/yolov7.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov7/aries/global/yolov7.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [640, 640],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "anchors": [
                [12, 16, 19, 36, 40, 28],
                [36, 75, 76, 55, 72, 146],
                [142, 110, 192, 243, 459, 401],
            ],
        },
    )
    DEFAULT = COCO_V1


class YOLOv7d6_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov7d6/aries/single/yolov7d6.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov7d6/aries/multi/yolov7d6.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov7d6/aries/global/yolov7d6.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [1280, 1280],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "anchors": [
                [19, 27, 44, 40, 38, 94],  # P3/8
                [96, 68, 86, 152, 180, 137],  # P4/16
                [140, 301, 303, 264, 238, 542],  # P5/32
                [436, 615, 739, 380, 925, 792],  # P6/64
            ],
        },
    )
    DEFAULT = COCO_V1


class YOLOv7e6_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov7e6/aries/single/yolov7e6.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov7e6/aries/multi/yolov7e6.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov7e6/aries/global/yolov7e6.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [1280, 1280],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "anchors": [
                [19, 27, 44, 40, 38, 94],  # P3/8
                [96, 68, 86, 152, 180, 137],  # P4/16
                [140, 301, 303, 264, 238, 542],  # P5/32
                [436, 615, 739, 380, 925, 792],  # P6/64
            ],
        },
    )
    DEFAULT = COCO_V1


class YOLOv7e6e_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov7e6e/aries/single/yolov7e6e.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov7e6e/aries/multi/yolov7e6e.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov7e6e/aries/global/yolov7e6e.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [1280, 1280],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "anchors": [
                [19, 27, 44, 40, 38, 94],  # P3/8
                [96, 68, 86, 152, 180, 137],  # P4/16
                [140, 301, 303, 264, 238, 542],  # P5/32
                [436, 615, 739, 380, 925, 792],  # P6/64
            ],
        },
    )
    DEFAULT = COCO_V1


class YOLOv7w6_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov7w6/aries/single/yolov7w6.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov7w6/aries/multi/yolov7w6.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov7w6/aries/global/yolov7w6.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [1280, 1280],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "anchors": [
                [19, 27, 44, 40, 38, 94],  # P3/8
                [96, 68, 86, 152, 180, 137],  # P4/16
                [140, 301, 303, 264, 238, 542],  # P5/32
                [436, 615, 739, 380, 925, 792],  # P6/64
            ],
        },
    )
    DEFAULT = COCO_V1


class YOLOv7x_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov7x/aries/single/yolov7x.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov7x/aries/multi/yolov7x.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov7x/aries/global/yolov7x.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [640, 640],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "anchors": [
                [12, 16, 19, 36, 40, 28],
                [36, 75, 76, 55, 72, 146],
                [142, 110, 192, 243, 459, 401],
            ],
        },
    )
    DEFAULT = COCO_V1


def YOLOv7(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv7_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def YOLOv7d6(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv7d6_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def YOLOv7e6(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv7e6_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def YOLOv7e6e(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv7e6e_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def YOLOv7w6(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv7w6_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def YOLOv7x(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv7x_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )



from abc import ABC, abstractmethod
import torch
from typing import Union, List


class PreOps(ABC):
    """Base class for preprocess operations."""

    def __init__(self):
        """Initialize the PreOps class."""
        super().__init__()
        self.device = torch.device("cpu")

    @abstractmethod
    def __call__(self, x):
        pass

    def to(self, device: Union[str, torch.device]):
        if isinstance(device, str):
            self.device = torch.device(device)
        elif isinstance(device, torch.device):
            self.device = device
        else:
            raise TypeError(f"Got unexpected type for device={type(device)}.")

        for name, value in self.__dict__.items():
            if isinstance(value, torch.Tensor):
                setattr(self, name, value.to(self.device))


class PreBase:
    """Base class for preprocess."""

    def __init__(self, Ops: List[PreOps]):
        """Initialize the PreBase class.

        Args:
            Ops (list): List of operations to be applied.
        """
        self.Ops = Ops
        self._check_ops()
        self.device = torch.device("cpu")

    def _check_ops(self):
        """Check if the operations are valid."""
        for op in self.Ops:
            if not isinstance(op, PreOps):
                raise TypeError(f"Got unsupported type={type(op)}.")

    def __call__(self, x):
        """Apply the operations to the input.

        Args:
            x: Input data.

        Returns:
            x: Processed data.
        """
        for op in self.Ops:
            x = op(x)
        return x

    def to(self, device: Union[str, torch.device]):
        """Move the operations to the specified device.

        Args:
            device (Union[str, torch.device]): Device to move the operations to.
        """
        if isinstance(device, str):
            self.device = torch.device(device)
        elif isinstance(device, torch.device):
            self.device = device
        else:
            raise TypeError(f"Got unexpected type for device={type(device)}.")

        for name, value in self.__dict__.items():
            if isinstance(value, torch.Tensor):
                setattr(self, name, value.to(self.device))

        for op in self.Ops:
            op.to(self.device)



from .alexnet import AlexNet
from .convnext import ConvNeXt_Tiny, ConvNeXt_Small, ConvNeXt_Base, ConvNeXt_Large
from .deit import (
    DeiT_Tiny_Patch16_224,
    DeiT_Small_Patch16_224,
    DeiT_Base_Patch16_224,
    DeiT_Base_Patch16_384,
)
from .deit3 import (
    DeiT3_Small_Patch16_224,
    DeiT3_Small_Patch16_384,
    DeiT3_Medium_Patch16_224,
    DeiT3_Base_Patch16_224,
    DeiT3_Base_Patch16_384,
    DeiT3_Large_Patch16_224,
    DeiT3_Large_Patch16_384,
)
from .densenet import DenseNet121, DenseNet169, DenseNet201
from .flexivit import FlexiViT_Small, FlexiViT_Base, FlexiViT_Large
from .googlenet import GoogLeNet
from .inception_v3 import Inception_V3
from .mnasnet import MNasNet1_0, MNasNet1_3
from .mobilenet_v2 import MobileNet_V2
from .regnet import (
    RegNet_X_16GF,
    RegNet_X_1_6GF,
    RegNet_X_32GF,
    RegNet_X_3_2GF,
    RegNet_X_400MF,
    RegNet_X_800MF,
    RegNet_X_8GF,
    RegNet_Y_16GF,
    RegNet_Y_1_6GF,
    RegNet_Y_32GF,
    RegNet_Y_3_2GF,
    RegNet_Y_400MF,
    RegNet_Y_800MF,
    RegNet_Y_8GF,
)
from .resnet import ResNet18, ResNet34, ResNet50, ResNet101, ResNet152
from .resnext import ResNext50_32x4d, ResNext101_64x4d, ResNext101_32x8d
from .shufflenet_v2 import (
    ShuffleNet_V2_X1_0,
    ShuffleNet_V2_X1_5,
    ShuffleNet_V2_X2_0,
)
from .vgg import VGG11, VGG11_BN, VGG13, VGG13_BN, VGG16, VGG16_BN, VGG19, VGG19_BN
from .vit import (
    ViT_Tiny_Patch16_224,
    ViT_Tiny_Patch16_384,
    ViT_Small_Patch16_224,
    ViT_Small_Patch16_384,
    ViT_Small_Patch32_224,
    ViT_Small_Patch32_384,
    ViT_Base_Patch8_224,
    ViT_Base_Patch16_224,
    ViT_Base_Patch16_384,
    ViT_Base_Patch32_224,
    ViT_Base_Patch32_384,
    ViT_Large_Patch16_224,
    ViT_Large_Patch16_384,
    ViT_Large_Patch32_384,
)
from .wide_resnet import Wide_ResNet50_2, Wide_ResNet101_2



from ..utils.types import ModelInfo, ModelInfoSet
from ..wrapper import MBLT_Engine


class YOLOv3u_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov3u/aries/single/yolov3u.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov3u/aries/multi/yolov3u.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov3u/aries/global/yolov3u.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [640, 640],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "nl": 3,  # Number of detection layers
        },
    )
    DEFAULT = COCO_V1


class YOLOv3_sppu_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov3-sppu/aries/single/yolov3-sppu.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov3-sppu/aries/multi/yolov3-sppu.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov3-sppu/aries/global/yolov3-sppu.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [640, 640],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "nl": 3,  # Number of detection layers
        },
    )
    DEFAULT = COCO_V1


class YOLOv3_spp_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov3-spp/aries/single/yolov3-spp.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov3-spp/aries/multi/yolov3-spp.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov3-spp/aries/global/yolov3-spp.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [640, 640],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "anchors": [
                [10, 13, 16, 30, 33, 23],  # P3/8
                [30, 61, 62, 45, 59, 119],  # P4/16
                [116, 90, 156, 198, 373, 326],  # P5/32
            ],
        },
    )
    DEFAULT = COCO_V1


def YOLOv3u(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv3u_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def YOLOv3_sppu(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv3_sppu_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def YOLOv3_spp(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv3_spp_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )



from .yolov3 import YOLOv3u, YOLOv3_sppu, YOLOv3_spp
from .yolov5 import (
    YOLOv5su,
    YOLOv5s6u,
    YOLOv5mu,
    YOLOv5m6u,
    YOLOv5lu,
    YOLOv5l6,
    YOLOv5l6u,
    YOLOv5xu,
    YOLOv5x6,
    YOLOv5x6u,
)
from .yolov7 import YOLOv7, YOLOv7d6, YOLOv7e6, YOLOv7e6e, YOLOv7w6, YOLOv7x
from .yolov8 import YOLOv8s, YOLOv8m, YOLOv8l, YOLOv8x
from .yolov9 import YOLOv9m, YOLOv9c, GELANm, GELANc
from .yolov10 import YOLOv10b
from .yolo11 import YOLO11s, YOLO11m, YOLO11l, YOLO11x
from .yolo12 import YOLO12s, YOLO12m



from .face_detection import *
from .image_classification import *
from .instance_segmentation import *
from .object_detection import *
from .pose_estimation import *

from ._api import list_models, list_tasks



from typing import Any, Dict, Optional, Tuple

import maccel
import torch

from transformers.cache_utils import CacheLayerMixin, Cache


class MobilintLayer(CacheLayerMixin):
    is_sliding = False
    
    def __init__(self, mxq_model: maccel.Model):
        self.mxq_model = mxq_model
        self._seen_tokens = 0
        
    def lazy_initialization(self, key_states: torch.Tensor):
        raise NotImplementedError("lazy_initialization is not implemented")

    def update(
        self, key_states: torch.Tensor, value_states: torch.Tensor, cache_kwargs: Optional[dict[str, Any]] = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        raise NotImplementedError("update is not implemented")

    def get_mask_sizes(self, cache_position: torch.Tensor) -> tuple[int, int]:
        kv_offset = 0
        query_length = cache_position.shape[0]
        past_seen_tokens = self.get_seq_length()
        kv_length = query_length + past_seen_tokens
        return kv_length, kv_offset

    def get_seq_length(self, cache_position=None) -> int:
        return self._seen_tokens

    def get_max_cache_shape(self) -> Optional[int]:
        return self.mxq_model.get_input_buffer_info()[0].max_cache_size
    
    def reset(self) -> None:
        self.mxq_model.reset_cache_memory()
        self._seen_tokens: int = 0

    def reorder_cache(self, beam_idx: torch.LongTensor):
        raise NotImplementedError("reorder_cache is not implemented")
    
    def update_cache_position(self, cache_position: torch.Tensor):
        self._seen_tokens += cache_position.numel()
    
class MobilintCache(Cache):
    def __init__(self, mxq_model: maccel.Model):
        self.mxq_model = mxq_model
        
        self.layers: list[MobilintLayer] = [MobilintLayer(self.mxq_model)]
        self.layer_classes = MobilintLayer
        
        self.num_hidden_layers = 1
        
        self.cache_processor = None
    
    def update_cache_position(self, cache_position: torch.Tensor):
        self.layers[0].update_cache_position(cache_position)


from typing import Optional, Tuple, TypeVar, Union
import math
import maccel
import torch
from torch import nn
from torch.nn import CrossEntropyLoss


from transformers import (
    WhisperTokenizer,
    WhisperFeatureExtractor,
    WhisperProcessor,
    PreTrainedModel,
    WhisperConfig,
    WhisperPreTrainedModel,
    AutoConfig,
    AutoTokenizer,
    AutoFeatureExtractor,
    AutoProcessor,
    AutoModelForSpeechSeq2Seq,
)
from transformers.models.whisper.modeling_whisper import (
    WhisperPositionalEmbedding,
    shift_tokens_right,
)
from transformers.models.whisper.generation_whisper import (
    WhisperGenerationMixin,
)
from transformers.modeling_outputs import (
    BaseModelOutput,
    BaseModelOutputWithPastAndCrossAttentions,
    Seq2SeqModelOutput,
    Seq2SeqLMOutput,
)
from transformers.utils import logging

from mblt_model_zoo.transformers.utils.generation_utils import MobilintGenerationMixin
from ..utils.cache_utils import MobilintCache


logger = logging.get_logger(__name__)

SpecificPreTrainedModelType = TypeVar(
    "SpecificPreTrainedModelType", bound="PreTrainedModel"
)


class MobilintWhisperConfig(WhisperConfig):
    model_type = "mobilint-whisper"

    def __init__(
        self,
        encoder_mxq_path: str = "",
        decoder_mxq_path: str = "",
        dev_no: int = 0,
        **kwargs,
    ):
        self.encoder_mxq_path = encoder_mxq_path
        self.decoder_mxq_path = decoder_mxq_path
        self.dev_no = dev_no

        super().__init__(**kwargs)

        self.tie_word_embeddings = False


class MobilintWhisperPreTrainedModel(WhisperPreTrainedModel):
    config_class = MobilintWhisperConfig
    supports_gradient_checkpointing = False
    _supports_flash_attn_2 = False
    _supports_sdpa = False
    _supports_static_cache = False

class Object(object):
    pass

class MobilintWhisperEncoder(MobilintWhisperPreTrainedModel):
    def __init__(self, config: MobilintWhisperConfig):
        super().__init__(config)

        self.dropout = config.dropout
        self.layerdrop = config.encoder_layerdrop

        embed_dim = config.d_model
        self.num_mel_bins = config.num_mel_bins
        self.padding_idx = config.pad_token_id
        self.max_source_positions = config.max_source_positions
        self.embed_scale = math.sqrt(embed_dim) if config.scale_embedding else 1.0
        
        self.conv1 = lambda: None
        self.conv2 = lambda: None
        
        self.conv1.stride = [1]
        self.conv2.stride = [2]

        self.gradient_checkpointing = False

        self.dev_no = config.dev_no
        self.acc = maccel.Accelerator(self.dev_no)
        mc = maccel.ModelConfig()
        mc.set_global_core_mode([maccel.Cluster.Cluster1])
        self.mxq_model = maccel.Model(
            f"{config.name_or_path}/{config.encoder_mxq_path}", mc
        )
        self.mxq_model.launch(self.acc)

    def _freeze_parameters(self):
        raise NotImplementedError("_freeze_parameters is not implemented")

    def get_input_embeddings(self) -> nn.Module:
        logger.warning_once("get_input_embeddings is not implemented")
        return None

    def set_input_embeddings(self, value: nn.Module):
        raise NotImplementedError("set_input_embeddings is not implemented")

    def forward(
        self,
        input_features,
        attention_mask=None,
        head_mask=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        expected_seq_length = (
            self.config.max_source_positions * self.conv1.stride[0] * self.conv2.stride[0]
        )
        if input_features.shape[-1] != expected_seq_length:
            raise ValueError(
                f"Whisper expects the mel input features to be of length {expected_seq_length}, but found {input_features.shape[-1]}. Make sure to pad the input mel features to {expected_seq_length}."
            )

        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        if head_mask is not None:
            logger.warning_once("head_mask is not supported.")

        if output_attentions:
            logger.warning_once("output_attentions is not supported.")

        if output_hidden_states:
            logger.warning_once("output_hidden_states is not supported.")

        output = self.mxq_model.infer(
            input_features.permute(0, 2, 1).type(torch.float32).cpu().numpy()
        )[0]
        hidden_states = torch.tensor(output, dtype=torch.float32, device=input_features.device).unsqueeze(0)

        if not return_dict:
            return (hidden_states,)
        return BaseModelOutput(
            last_hidden_state=hidden_states, hidden_states=None, attentions=None
        )

    def dispose(self):
        self.mxq_model.dispose()


class MobilintWhisperDecoder(MobilintWhisperPreTrainedModel):
    main_input_name = "input_ids"

    def __init__(self, config: MobilintWhisperConfig):
        super().__init__(config)
        self.dropout = config.dropout
        self.layerdrop = config.decoder_layerdrop
        self.padding_idx = config.pad_token_id
        self.max_target_positions = config.max_target_positions
        self.max_source_positions = config.max_source_positions
        self.embed_scale = math.sqrt(config.d_model) if config.scale_embedding else 1.0

        self.embed_tokens = nn.Embedding(
            config.vocab_size, config.d_model, self.padding_idx
        )
        self.embed_positions = WhisperPositionalEmbedding(
            self.max_target_positions, config.d_model
        )

        self._use_flash_attention_2 = False
        self._use_sdpa = False

        self.gradient_checkpointing = False
        # Initialize weights and apply final processing
        self.post_init()

        self.dev_no = config.dev_no
        self.acc = maccel.Accelerator(self.dev_no)
        mc = maccel.ModelConfig()
        mc.set_single_core_mode(
            None, [maccel.CoreId(maccel.Cluster.Cluster0, maccel.Core.Core3)]
        )
        self.mxq_model = maccel.Model(
            f"{config.name_or_path}/{config.decoder_mxq_path}", mc
        )
        self.mxq_model.launch(self.acc)

    def get_input_embeddings(self):
        return self.embed_tokens

    def set_input_embeddings(self, value):
        self.embed_tokens = value

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        encoder_hidden_states=None,
        head_mask=None,
        cross_attn_head_mask=None,
        past_key_values: Optional[MobilintCache] = None,
        inputs_embeds=None,
        position_ids=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        cache_position=None,
    ):
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        use_cache = use_cache if use_cache is not None else self.config.use_cache
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        if attention_mask is not None:
            logger.warning_once("attention_mask is not supported.")

        if head_mask is not None:
            logger.warning_once("head_mask is not supported.")

        if cross_attn_head_mask is not None:
            logger.warning_once("cross_attn_head_mask is not supported.")

        if output_attentions:
            logger.warning_once("output_attentions is not supported.")

        if output_hidden_states:
            logger.warning_once("output_hidden_states is not supported.")

        # retrieve input_ids and inputs_embeds
        if input_ids is not None and inputs_embeds is not None:
            raise ValueError(
                "You cannot specify both decoder_input_ids and decoder_inputs_embeds at the same time"
            )
        elif input_ids is not None:
            input_shape = input_ids.size()
            input_ids = input_ids.view(-1, input_shape[-1])
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
        else:
            raise ValueError(
                "You have to specify either decoder_input_ids or decoder_inputs_embeds"
            )

        if inputs_embeds is None:
            inputs_embeds = self.embed_tokens(input_ids)

        if use_cache or past_key_values is not None:
            if not isinstance(past_key_values, MobilintCache):
                logger.warning_once(
                    "Class of past_key_values should be MobilintCache, current: "
                    + past_key_values.__class__.__name__
                )
                past_key_values = MobilintCache(self.mxq_model)

        past_key_values_length = 0
        if cache_position is not None:
            past_key_values_length = cache_position[0]
        elif past_key_values is not None:
            past_key_values_length = past_key_values.get_seq_length()

        if cache_position is None:
            cache_position = torch.arange(
                past_key_values_length,
                past_key_values_length + input_shape[1],
                device=inputs_embeds.device,
            )

        if position_ids is None:
            position_ids = cache_position.unsqueeze(0).repeat(input_shape[0], 1)

        # embed positions
        if input_ids is not None:
            positions = self.embed_positions(
                input_ids,
                past_key_values_length=past_key_values_length,
                position_ids=position_ids,
            )
        else:
            positions = self.embed_positions(
                inputs_embeds,
                past_key_values_length=past_key_values_length,
                position_ids=position_ids,
            )

        hidden_states = inputs_embeds + positions.to(inputs_embeds.device)

        inputs = [
            encoder_hidden_states.type(torch.float32).cpu().numpy(),
            hidden_states.unsqueeze(1).type(torch.float32).cpu().numpy(),
        ]
        logits = self.mxq_model.infer(inputs, cache_size=int(past_key_values_length))[0]
        logits = torch.tensor(logits, dtype=torch.float32, device=self.device)

        if use_cache:
            past_key_values.update_cache_position(cache_position)

        next_cache = past_key_values if use_cache else None
        if not return_dict:
            return tuple(logits, next_cache)
        return BaseModelOutputWithPastAndCrossAttentions(
            last_hidden_state=logits,
            past_key_values=next_cache,
            hidden_states=None,
            attentions=None,
            cross_attentions=None,
        )

    def dispose(self):
        self.mxq_model.dispose()


class MobilintWhisperModel(MobilintWhisperPreTrainedModel):
    def __init__(self, config: WhisperConfig):
        super().__init__(config)

        self.encoder = MobilintWhisperEncoder(config)
        self.decoder = MobilintWhisperDecoder(config)

    def get_input_embeddings(self):
        return self.decoder.embed_tokens

    def set_input_embeddings(self, value):
        self.decoder.embed_tokens = value

    def get_encoder(self):
        return self.encoder

    def get_decoder(self):
        return self.decoder

    def freeze_encoder(self):
        self.encoder._freeze_parameters()

    def forward(
        self,
        input_features: Optional[torch.FloatTensor] = None,
        attention_mask: Optional[torch.LongTensor] = None,
        decoder_input_ids: Optional[torch.LongTensor] = None,
        decoder_attention_mask: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        decoder_head_mask: Optional[torch.Tensor] = None,
        cross_attn_head_mask: Optional[torch.Tensor] = None,
        encoder_outputs: Optional[Tuple[Tuple[torch.FloatTensor]]] = None,
        past_key_values: Optional[MobilintCache] = None,
        decoder_inputs_embeds: Optional[Tuple[torch.FloatTensor]] = None,
        decoder_position_ids: Optional[Tuple[torch.LongTensor]] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        cache_position: Optional[torch.LongTensor] = None,
    ) -> Union[Tuple[torch.Tensor], Seq2SeqModelOutput]:
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        use_cache = use_cache if use_cache is not None else self.config.use_cache
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        if encoder_outputs is None:
            if attention_mask is not None:
                logger.warning_once("attention_mask is not supported.")

            encoder_outputs = self.encoder(
                input_features,
                head_mask=head_mask,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                return_dict=return_dict,
            )
        # If the user passed a tuple for encoder_outputs, we wrap it in a BaseModelOutput when return_dict=True
        elif return_dict and not isinstance(encoder_outputs, BaseModelOutput):
            encoder_outputs = BaseModelOutput(
                last_hidden_state=encoder_outputs[0],
                hidden_states=encoder_outputs[1] if len(encoder_outputs) > 1 else None,
                attentions=encoder_outputs[2] if len(encoder_outputs) > 2 else None,
            )

        # decoder outputs consists of (dec_features, past_key_value, dec_hidden, dec_attn)
        decoder_outputs = self.decoder(
            input_ids=decoder_input_ids,
            attention_mask=decoder_attention_mask,
            encoder_hidden_states=encoder_outputs[0],
            head_mask=decoder_head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            past_key_values=past_key_values,
            inputs_embeds=decoder_inputs_embeds,
            position_ids=decoder_position_ids,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            cache_position=cache_position,
        )

        if not return_dict:
            return decoder_outputs + encoder_outputs

        return Seq2SeqModelOutput(
            last_hidden_state=decoder_outputs.last_hidden_state,
            past_key_values=decoder_outputs.past_key_values,
            decoder_hidden_states=decoder_outputs.hidden_states,
            decoder_attentions=decoder_outputs.attentions,
            cross_attentions=decoder_outputs.cross_attentions,
            encoder_last_hidden_state=encoder_outputs.last_hidden_state,
            encoder_hidden_states=encoder_outputs.hidden_states,
            encoder_attentions=encoder_outputs.attentions,
        )

    def dispose(self):
        self.encoder.dispose()
        self.decoder.dispose()


class MobilintWhisperForConditionalGeneration(
    MobilintGenerationMixin, WhisperGenerationMixin, MobilintWhisperPreTrainedModel
):
    base_model_prefix = "model"

    def __init__(self, config: MobilintWhisperConfig, *inputs, **kwargs):
        super().__init__(config, *inputs, **kwargs)
        self.model = MobilintWhisperModel(config)
        self.max_target_positions = config.max_target_positions
        # for pipeline type checking
        self.config.model_type = "whisper"
    
    def get_mxq_model(self):
        return self.model.decoder.mxq_model

    def get_encoder(self):
        return self.model.get_encoder()

    def get_decoder(self):
        return self.model.get_decoder()

    def get_output_embeddings(self):
        logger.warning_once("get_output_embeddings is not implemented")
        return None

    def set_output_embeddings(self, new_embeddings):
        raise NotImplementedError("set_output_embeddings is not implemented")

    def get_input_embeddings(self) -> nn.Module:
        return self.model.get_input_embeddings()

    def freeze_encoder(self):
        self.model.encoder._freeze_parameters()

    def tie_weights(self):
        pass

    def forward(
        self,
        input_features: Optional[torch.FloatTensor] = None,
        attention_mask: Optional[torch.LongTensor] = None,
        decoder_input_ids: Optional[torch.LongTensor] = None,
        decoder_attention_mask: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        decoder_head_mask: Optional[torch.Tensor] = None,
        cross_attn_head_mask: Optional[torch.Tensor] = None,
        encoder_outputs: Optional[Tuple[Tuple[torch.FloatTensor]]] = None,
        past_key_values: Optional[MobilintCache] = None,
        decoder_inputs_embeds: Optional[Tuple[torch.FloatTensor]] = None,
        decoder_position_ids: Optional[Tuple[torch.LongTensor]] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        cache_position: Optional[torch.LongTensor] = None,
    ) -> Union[Tuple[torch.Tensor], Seq2SeqLMOutput]:
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        if labels is not None:
            if labels.shape[1] > self.max_target_positions:
                raise ValueError(
                    f"Labels' sequence length {labels.shape[1]} cannot exceed the maximum allowed length of {self.max_target_positions} tokens."
                )
            if decoder_input_ids is None and decoder_inputs_embeds is None:
                decoder_input_ids = shift_tokens_right(
                    labels, self.config.pad_token_id, self.config.decoder_start_token_id
                )

        outputs = self.model(
            input_features,
            attention_mask=attention_mask,
            decoder_input_ids=decoder_input_ids,
            encoder_outputs=encoder_outputs,
            decoder_attention_mask=decoder_attention_mask,
            head_mask=head_mask,
            decoder_head_mask=decoder_head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            past_key_values=past_key_values,
            decoder_inputs_embeds=decoder_inputs_embeds,
            decoder_position_ids=decoder_position_ids,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            cache_position=cache_position,
        )
        lm_logits = outputs[0].squeeze(1)  # proj_out is performed on decoder mblt.

        loss = None
        if labels is not None:
            loss_fct = CrossEntropyLoss()
            # move labels to correct device to enable PP
            labels = labels.to(lm_logits.device)
            loss = loss_fct(
                lm_logits.view(-1, self.config.vocab_size), labels.reshape(-1)
            )

        if not return_dict:
            output = (lm_logits,) + outputs[1:]
            return ((loss,) + output) if loss is not None else output

        return Seq2SeqLMOutput(
            loss=loss,
            logits=lm_logits,
            past_key_values=outputs.past_key_values,
            decoder_hidden_states=outputs.decoder_hidden_states,
            decoder_attentions=outputs.decoder_attentions,
            cross_attentions=outputs.cross_attentions,
            encoder_last_hidden_state=outputs.encoder_last_hidden_state,
            encoder_hidden_states=outputs.encoder_hidden_states,
            encoder_attentions=outputs.encoder_attentions,
        )
    
    def dispose(self):
        self.model.dispose()


AutoConfig.register("mobilint-whisper", MobilintWhisperConfig)
AutoTokenizer.register(MobilintWhisperConfig, WhisperTokenizer)
AutoFeatureExtractor.register(MobilintWhisperConfig, WhisperFeatureExtractor)
AutoProcessor.register(MobilintWhisperConfig, WhisperProcessor)
AutoModelForSpeechSeq2Seq.register(
    MobilintWhisperConfig, MobilintWhisperForConditionalGeneration
)

from ..utils.types import TransformersModelInfo

whisper_small = TransformersModelInfo(
    original_model_id="openai/whisper-small",
    model_id="mobilint/whisper-small",
    download_url_base="https://dl.mobilint.com/model/transformers/stt/whisper-small/",
    file_list=[
        "added_tokens.json",
        "config.json",
        "generation_config.json",
        "merges.txt",
        "model.safetensors",
        "normalizer.json",
        "preprocessor_config.json",
        "special_tokens_map.json",
        "tokenizer.json",
        "tokenizer_config.json",
        "vocab.json",
        "whisper-small_encoder.mxq",
        "whisper-small_decoder.mxq",
    ],
)



from ..utils.types import ModelInfo, ModelInfoSet
from ..wrapper import MBLT_Engine


class DenseNet121_Set(ModelInfoSet):
    IMAGENET1K_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/densenet121_IMAGENET1K_V1/aries/single/densenet121_IMAGENET1K_V1.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/densenet121_IMAGENET1K_V1/aries/multi/densenet121_IMAGENET1K_V1.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/densenet121_IMAGENET1K_V1/aries/global/densenet121_IMAGENET1K_V1.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 256,
                "interpolation": "bilinear",
            },
            "CenterCrop": {
                "size": [224, 224],
            },
            "Normalize": {"style": "torch"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )
    DEFAULT = IMAGENET1K_V1  # Default model


class DenseNet169_Set(ModelInfoSet):
    IMAGENET1K_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/densenet169_IMAGENET1K_V1/aries/single/densenet169_IMAGENET1K_V1.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/densenet169_IMAGENET1K_V1/aries/multi/densenet169_IMAGENET1K_V1.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/densenet169_IMAGENET1K_V1/aries/global/densenet169_IMAGENET1K_V1.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 256,
                "interpolation": "bilinear",
            },
            "CenterCrop": {
                "size": [224, 224],
            },
            "Normalize": {"style": "torch"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )
    DEFAULT = IMAGENET1K_V1  # Default model


class DenseNet201_Set(ModelInfoSet):
    IMAGENET1K_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/densenet201_IMAGENET1K_V1/aries/single/densenet201_IMAGENET1K_V1.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/densenet201_IMAGENET1K_V1/aries/multi/densenet201_IMAGENET1K_V1.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/densenet201_IMAGENET1K_V1/aries/global/densenet201_IMAGENET1K_V1.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 256,
                "interpolation": "bilinear",
            },
            "CenterCrop": {
                "size": [224, 224],
            },
            "Normalize": {"style": "torch"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )
    DEFAULT = IMAGENET1K_V1  # Default model


def DenseNet121(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        DenseNet121_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def DenseNet169(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        DenseNet169_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def DenseNet201(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        DenseNet201_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )



import numpy as np
import torch
from PIL import Image
from typing import Union
from .base import PreOps
from ..types import TensorLike


class Normalize(PreOps):
    def __init__(
        self,
        style: str = "torch",
    ):
        """Normalize image

        Args:
            style (str): Normalization style. Supported values are "torch", "tf"

        """
        super().__init__()
        self.style = style
        if self.style not in ["torch", "tf"]:
            raise ValueError(f"style {self.style} not supported.")

    def __call__(self, x: Union[TensorLike, Image.Image]):
        """Normalize image

        Args:
            x (np.ndarray): Image to be normalized.

        Returns:
            x: Normalized image.
        """
        if isinstance(x, np.ndarray):
            x = torch.from_numpy(x).float().to(self.device)
        elif isinstance(x, Image.Image):
            x = torch.from_numpy(np.array(x)).float().to(self.device)
        elif isinstance(x, torch.Tensor):
            x = x.to(self.device)
        else:
            raise TypeError(f"Got unexpected type for x={type(x)})")

        if self.style == "torch":
            x = x / 255.0
            x -= torch.from_numpy(np.array([0.485, 0.456, 0.406])).to(self.device)
            x /= torch.from_numpy(np.array([0.229, 0.224, 0.225])).to(self.device)
        elif self.style == "tf":
            x = x / 127.5 - 1.0
        return x



from abc import ABC, abstractmethod
from typing import Union
import numpy as np
import torch
from ..types import TensorLike, ListTensorLike
from .common import *


class PostBase(ABC):
    def __init__(self):
        super().__init__()
        self.device = torch.device("cpu")

    @abstractmethod
    def __call__(self, x):
        pass

    def to(self, device: Union[str, torch.device]):
        """Move the operations to the specified device.

        Args:
            device (Union[str, torch.device]): Device to move the operations to.
        """
        if isinstance(device, str):
            self.device = torch.device(device)
        elif isinstance(device, torch.device):
            self.device = device
        else:
            raise TypeError(f"Got unexpected type for device={type(device)}.")

        for name, value in self.__dict__.items():
            if isinstance(value, torch.Tensor):
                setattr(self, name, value.to(self.device))


class YOLOPostBase(PostBase):
    def __init__(self, pre_cfg: dict, post_cfg: dict):
        super().__init__()
        img_size = pre_cfg.get("YoloPre")["img_size"]

        if isinstance(img_size, int):
            self.imh = self.imw = img_size
        elif isinstance(img_size, list):
            self.imh, self.imw = img_size

        self.nc = post_cfg.get("nc")
        self.anchors = post_cfg.get("anchors", None)  # anchor coordinates
        if self.anchors is None:
            self.anchorless = True
            self.nl = post_cfg.get("nl")
            assert self.nl is not None, "nl should be provided for anchorless model"
        else:
            self.anchorless = False
            self.nl = len(self.anchors)
            self.na = len(self.anchors[0]) // 2

        self.n_extra = post_cfg.get("n_extra", 0)
        self.task = post_cfg.get("task")

    def __call__(self, x, conf_thres=None, iou_thres=None):
        self.set_threshold(conf_thres, iou_thres)
        x = self.check_input(x)
        x = self.rearrange(x)
        x = self.decode(x)
        x = self.nms(x)
        return x

    def set_threshold(self, conf_thres: float = None, iou_thres: float = None):
        assert (
            conf_thres is not None and iou_thres is not None
        ), "conf_thres and iou_thres should be provided in yolo_postprocess "
        assert 0 < conf_thres < 1, "conf_thres should be in (0, 1)"
        assert 0 < iou_thres < 1, "iou_thres should be in (0, 1)"
        self.conf_thres = conf_thres
        self.inv_conf_thres = -np.log(1 / conf_thres - 1)
        self.iou_thres = iou_thres

    def check_input(self, x: Union[TensorLike, ListTensorLike]):
        if isinstance(x, np.ndarray):
            x = [torch.from_numpy(x)]
        elif isinstance(x, torch.Tensor):
            x = [x]

        assert isinstance(x, list), f"Got unexpected type for x={type(x)}."

        if isinstance(x[0], np.ndarray):
            x = [torch.from_numpy(xi).to(self.device) for xi in x]
        elif isinstance(x[0], torch.Tensor):
            x = [xi.to(self.device) for xi in x]

        return x

    @abstractmethod
    def rearrange(self, x):
        pass

    @abstractmethod
    def decode(self, x):
        pass

    @abstractmethod
    def nms(self, x):
        pass

    def masking(self, x, proto_outs):
        masks = []
        for pred, proto in zip(x, proto_outs):
            if len(pred) == 0:
                masks.append(
                    torch.zeros(
                        (0, self.imh, self.imw), dtype=torch.float32, device=self.device
                    )
                )
                continue
            masks.append(
                process_mask_upsample(
                    proto, pred[:, 6:], pred[:, :4], [self.imh, self.imw]
                )
            )
        return [[xi, mask] for xi, mask in zip(x, masks)]



from typing import Optional, Tuple, TypeVar, Union

import maccel
import torch
import torch.nn as nn
import numpy as np
import math
from torch.nn import CrossEntropyLoss

from transformers import (
    PretrainedConfig,
    PreTrainedModel,
    GPT2Tokenizer,
    GPT2TokenizerFast,
    AutoConfig,
    AutoModel,
    AutoTokenizer,
    AutoModelForCausalLM,
)
from transformers.modeling_outputs import (
    CausalLMOutputWithPast,
    BaseModelOutputWithPast,
)
from transformers.utils import logging

from mblt_model_zoo.transformers.utils.generation_utils import MobilintGenerationMixin
from ..utils.cache_utils import MobilintCache


logger = logging.get_logger(__name__)

SpecificPreTrainedModelType = TypeVar(
    "SpecificPreTrainedModelType", bound="PreTrainedModel"
)


class ExaoneConfig(PretrainedConfig):
    model_type = "exaone"
    attribute_map = {"num_hidden_layers": "num_layers"}

    def __init__(
        self,
        vocab_size=102400,
        max_position_embeddings=2048,
        hidden_size=2048,
        num_layers=32,
        num_attention_heads=32,
        num_key_value_heads=None,
        intermediate_size=None,
        activation_function="silu",
        rope_theta=10000.0,
        rope_scaling=None,
        embed_dropout=0.0,
        attention_dropout=0.0,
        layer_norm_epsilon=1e-5,
        initializer_range=0.02,
        use_cache=True,
        bos_token_id=0,
        eos_token_id=2,
        **kwargs,
    ):
        self.vocab_size = vocab_size
        self.max_position_embeddings = max_position_embeddings
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.num_attention_heads = num_attention_heads
        self.num_layers = num_layers
        if num_key_value_heads is None:
            num_key_value_heads = num_attention_heads
        self.num_key_value_heads = num_key_value_heads
        if intermediate_size:
            self.intermediate_size = intermediate_size
        else:
            self.intermediate_size = hidden_size * 4
        self.activation_function = activation_function
        self.embed_dropout = embed_dropout
        self.attention_dropout = attention_dropout
        self.layer_norm_epsilon = layer_norm_epsilon
        self.initializer_range = initializer_range
        self.use_cache = use_cache
        self.rope_theta = rope_theta
        self.rope_scaling = rope_scaling

        self.bos_token_id = bos_token_id
        self.eos_token_id = eos_token_id

        super().__init__(bos_token_id=bos_token_id, eos_token_id=eos_token_id, **kwargs)


class MobilintExaoneConfig(ExaoneConfig):
    model_type = "mobilint-exaone"

    def __init__(
        self,
        mxq_path: str = "",
        dev_no: int = 0,
        **kwargs,
    ):
        self.mxq_path = mxq_path
        self.dev_no = dev_no

        super().__init__(**kwargs)

        self.tie_word_embeddings = False


class MobilintExaoneForCausalLM(PreTrainedModel, MobilintGenerationMixin):
    supports_gradient_checkpointing = False
    _supports_flash_attn_2 = False
    _supports_sdpa = False

    config_class = MobilintExaoneConfig

    def __init__(self, config: MobilintExaoneConfig, *inputs, **kwargs):
        super().__init__(config, *inputs, **kwargs)
        self.config = config
        self.embed_dim = config.hidden_size
        self.wte = nn.Embedding(
            config.vocab_size, self.embed_dim, self.config.pad_token_id
        )

        self.dev_no = config.dev_no
        self.acc = maccel.Accelerator(self.dev_no)
        mc = maccel.ModelConfig()
        mc.set_single_core_mode(1)
        self.mxq_model = maccel.Model(f"{config.name_or_path}/{config.mxq_path}", mc)
        self.mxq_model.launch(self.acc)
    
    def get_mxq_model(self):
        return self.mxq_model

    def get_output_embeddings(self):
        logger.warning_once("self.lm_head is implemented in mxq")
        return None

    def set_output_embeddings(self, new_embeddings):
        raise NotImplementedError("self.lm_head is implemented in mxq")

    def forward(
        self,
        input_ids: torch.LongTensor = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[MobilintCache] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        cache_position: Optional[torch.LongTensor] = None,
        chunk_size: int = 0,
    ) -> Union[Tuple[torch.Tensor], BaseModelOutputWithPast]:
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        use_cache = use_cache if use_cache is not None else self.config.use_cache
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        if output_attentions:
            logger.warning_once("output_attentions is not supported.")

        if output_hidden_states:
            logger.warning_once("output_hidden_states is not supported.")

        if (input_ids is None) ^ (inputs_embeds is not None):
            raise ValueError(
                "You must specify exactly one of input_ids or inputs_embeds"
            )

        if inputs_embeds is None:
            inputs_embeds = self.wte(input_ids)

        if use_cache and past_key_values is None:
            past_key_values = MobilintCache(self.mxq_model)

        if cache_position is None:
            past_seen_tokens = (
                past_key_values.get_seq_length() if past_key_values is not None else 0
            )
            cache_position = torch.arange(
                past_seen_tokens,
                past_seen_tokens + inputs_embeds.shape[1],
                dtype=torch.long,
                device=inputs_embeds.device,
            )

        inputs_embeds = inputs_embeds.type(torch.float32).cpu().numpy()

        if inputs_embeds.ndim == 3:
            inputs_embeds = np.expand_dims(
                inputs_embeds, 1
            )  # (batch, 1, seqlen, hidden_size)

        # max width should be appropriate number for chunking (ex. 192 for Llama 3.2 3B)
        # it should be searched experimentally
        if chunk_size == 0:
            chunk_size = self.mxq_model.get_input_buffer_info()[0].max_width
        num_of_chunks = math.ceil(inputs_embeds.shape[2] / chunk_size)

        for i in range(num_of_chunks):
            start_index = i * chunk_size
            end_index = min(start_index + chunk_size, inputs_embeds.shape[2])
            cache_size = (
                0 if past_key_values is None else past_key_values.get_seq_length()
            )

            # last infer
            if i == num_of_chunks - 1:
                logits = self.mxq_model.infer(
                    [inputs_embeds[:, :, start_index:end_index, :]], None, cache_size
                )[0]
            else:
                logits = self.mxq_model.infer(
                    [inputs_embeds[:, :, start_index:end_index, :]], None, cache_size
                )[0]

            if use_cache:
                past_key_values.update_cache_position(
                    cache_position[start_index:end_index]
                )

        lm_logits = torch.tensor(logits, dtype=torch.float32, device=self.device).squeeze(0)

        loss = None
        if labels is not None:
            lm_logits = lm_logits.to(torch.float32)

            # Shift so that tokens < n predict n
            shift_logits = lm_logits[..., :-1, :].contiguous()
            shift_labels = labels[..., 1:].contiguous()
            # Flatten the tokens
            loss_fct = CrossEntropyLoss()
            loss = loss_fct(
                shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1)
            )

            lm_logits = lm_logits.to(self.config.torch_dtype)
            loss = loss.to(self.config.torch_dtype)

        if not return_dict:
            output = (lm_logits,)
            return ((loss,) + output) if loss is not None else output

        return CausalLMOutputWithPast(
            loss=loss,
            logits=lm_logits,
            past_key_values=past_key_values if use_cache else None,
            hidden_states=None,
            attentions=None,
        )

    def dispose(self):
        self.mxq_model.dispose()


AutoConfig.register("mobilint-exaone", MobilintExaoneConfig)
AutoModel.register(MobilintExaoneConfig, MobilintExaoneForCausalLM)
AutoTokenizer.register(
    MobilintExaoneConfig,
    fast_tokenizer_class=GPT2TokenizerFast,
    slow_tokenizer_class=GPT2Tokenizer,
)
AutoModelForCausalLM.register(MobilintExaoneConfig, MobilintExaoneForCausalLM)

from ..utils.types import TransformersModelInfo

EXAONE_35_24B_Instruct = TransformersModelInfo(
    original_model_id="LGAI-EXAONE/EXAONE-3.5-2.4B-Instruct",
    model_id="mobilint/EXAONE-3.5-2.4B-Instruct",
    download_url_base="https://dl.mobilint.com/model/transformers/llm/EXAONE-3.5-2.4B-Instruct/",
    file_list=[
        "config.json",
        "EXAONE-3.5-2.4B-Instruct.mxq",
        "generation_config.json",
        "merges.txt",
        "model.safetensors",
        "special_tokens_map.json",
        "tokenizer.json",
        "tokenizer_config.json",
        "vocab.json",
    ],
)

EXAONE_Deep_24B = TransformersModelInfo(
    original_model_id="LGAI-EXAONE/EXAONE-Deep-2.4B",
    model_id="mobilint/EXAONE-Deep-2.4B",
    download_url_base="https://dl.mobilint.com/model/transformers/llm/EXAONE-Deep-2.4B/",
    file_list=[
        "config.json",
        "EXAONE-Deep-2.4B.mxq",
        "generation_config.json",
        "merges.txt",
        "model.safetensors",
        "special_tokens_map.json",
        "tokenizer.json",
        "tokenizer_config.json",
        "vocab.json",
    ],
)



from typing import List
from dataclasses import dataclass


@dataclass
class TransformersModelInfo:
    """
    This class is used to store model information from huggingface transformers library.
    """

    original_model_id: str
    model_id: str
    download_url_base: str
    file_list: List[str]

    def __str__(self):
        return f"'{self.model_id}'"

    def __repr__(self):
        return self.__str__()

    def get_directory_name(self):
        return self.model_id[len("mobilint/") :]



from ..utils.types import ModelInfo, ModelInfoSet
from ..wrapper import MBLT_Engine


class ViT_Tiny_Patch16_224_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/vit_tiny_patch16_224/aries/single/vit_tiny_patch16_224.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/vit_tiny_patch16_224/aries/multi/vit_tiny_patch16_224.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/vit_tiny_patch16_224/aries/global/vit_tiny_patch16_224.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 248,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [224, 224],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


class ViT_Tiny_Patch16_384_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/vit_tiny_patch16_384/aries/single/vit_tiny_patch16_384.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/vit_tiny_patch16_384/aries/multi/vit_tiny_patch16_384.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/vit_tiny_patch16_384/aries/global/vit_tiny_patch16_384.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 384,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [384, 384],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


class ViT_Small_Patch16_224_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/vit_small_patch16_224/aries/single/vit_small_patch16_224.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/vit_small_patch16_224/aries/multi/vit_small_patch16_224.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/vit_small_patch16_224/aries/global/vit_small_patch16_224.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 248,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [224, 224],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


class ViT_Small_Patch16_384_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/vit_small_patch16_384/aries/single/vit_small_patch16_384.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/vit_small_patch16_384/aries/multi/vit_small_patch16_384.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/vit_small_patch16_384/aries/global/vit_small_patch16_384.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 384,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [384, 384],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


class ViT_Small_Patch32_224_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/vit_small_patch32_224/aries/single/vit_small_patch32_224.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/vit_small_patch32_224/aries/multi/vit_small_patch32_224.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/vit_small_patch32_224/aries/global/vit_small_patch32_224.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 248,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [224, 224],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


class ViT_Small_Patch32_384_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/vit_small_patch32_384/aries/single/vit_small_patch32_384.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/vit_small_patch32_384/aries/multi/vit_small_patch32_384.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/vit_small_patch32_384/aries/global/vit_small_patch32_384.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 384,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [384, 384],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


class ViT_Base_Patch8_224_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/vit_base_patch8_224/aries/single/vit_base_patch8_224.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/vit_base_patch8_224/aries/multi/vit_base_patch8_224.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/vit_base_patch8_224/aries/global/vit_base_patch8_224.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 248,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [224, 224],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


class ViT_Base_Patch16_224_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/vit_base_patch16_224/aries/single/vit_base_patch16_224.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/vit_base_patch16_224/aries/multi/vit_base_patch16_224.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/vit_base_patch16_224/aries/global/vit_base_patch16_224.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 248,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [224, 224],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


class ViT_Base_Patch16_384_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/vit_base_patch16_384/aries/single/vit_base_patch16_384.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/vit_base_patch16_384/aries/multi/vit_base_patch16_384.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/vit_base_patch16_384/aries/global/vit_base_patch16_384.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 384,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [384, 384],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


class ViT_Base_Patch32_224_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/vit_base_patch32_224/aries/single/vit_base_patch32_224.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/vit_base_patch32_224/aries/multi/vit_base_patch32_224.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/vit_base_patch32_224/aries/global/vit_base_patch32_224.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 248,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [224, 224],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


class ViT_Base_Patch32_384_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/vit_base_patch32_384/aries/single/vit_base_patch32_384.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/vit_base_patch32_384/aries/multi/vit_base_patch32_384.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/vit_base_patch32_384/aries/global/vit_base_patch32_384.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 384,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [384, 384],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


class ViT_Large_Patch16_224_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/vit_large_patch16_224/aries/single/vit_large_patch16_224.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/vit_large_patch16_224/aries/multi/vit_large_patch16_224.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/vit_large_patch16_224/aries/global/vit_large_patch16_224.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 248,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [224, 224],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


class ViT_Large_Patch16_384_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/vit_large_patch16_384/aries/single/vit_large_patch16_384.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/vit_large_patch16_384/aries/multi/vit_large_patch16_384.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/vit_large_patch16_384/aries/global/vit_large_patch16_384.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 384,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [384, 384],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


class ViT_Large_Patch32_384_Set(ModelInfoSet):
    DEFAULT = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/image_classification/vit_large_patch32_384/aries/single/vit_large_patch32_384.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/image_classification/vit_large_patch32_384/aries/multi/vit_large_patch32_384.mxq",
                    "global": "https://dl.mobilint.com/model/vision/image_classification/vit_large_patch32_384/aries/global/vit_large_patch32_384.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "pil",
            },
            "Resize": {
                "size": 384,
                "interpolation": "bicubic",
            },
            "CenterCrop": {
                "size": [384, 384],
            },
            "Normalize": {"style": "tf"},
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={"task": "image_classification"},
    )


def ViT_Tiny_Patch16_224(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        ViT_Tiny_Patch16_224_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def ViT_Tiny_Patch16_384(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        ViT_Tiny_Patch16_384_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def ViT_Small_Patch16_224(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        ViT_Small_Patch16_224_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def ViT_Small_Patch16_384(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        ViT_Small_Patch16_384_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def ViT_Small_Patch32_224(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        ViT_Small_Patch32_224_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def ViT_Small_Patch32_384(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        ViT_Small_Patch32_384_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def ViT_Base_Patch8_224(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        ViT_Base_Patch8_224_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def ViT_Base_Patch16_224(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        ViT_Base_Patch16_224_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def ViT_Base_Patch16_384(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        ViT_Base_Patch16_384_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def ViT_Base_Patch32_224(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        ViT_Base_Patch32_224_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def ViT_Base_Patch32_384(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        ViT_Base_Patch32_384_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def ViT_Large_Patch16_224(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        ViT_Large_Patch16_224_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def ViT_Large_Patch16_384(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        ViT_Large_Patch16_384_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def ViT_Large_Patch32_384(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        ViT_Large_Patch32_384_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )
